#   -*- coding: utf-8 -*-
import re
import sys
import math
import logging
from colorama import init as colorama_init
from colorama import Style
from colorama import Fore
from colorama import Back
from colorama import Cursor

logger = logging.getLogger(__name__)

MAX_LINES = 75
MAX_CHARS = 120
HIDE_CURSOR = '\x1b[?25l'
SHOW_CURSOR = '\x1b[?25h'
CLEAR_EOL = '\033[K'
PROGRESS_TICKER = chr(9632)  # ■
PROGRESS_BAR_WIDTH = 50
ID_WIDTH = 30


class Terminal():
    """ simple ANSI terminal
        represented internally as a list of dictionaries
        ability to move up or down and print text at different offsets
    """
    def __init__(self, lines, config=None, create=True):
        """ class constructor
        """
        logger.debug('executing Terminal constructor')
        self.validate_lines(lines)
        self.lines = lines
        self.zfill = len(str(lines))

        if not config:
            config = {}
        self.validate_config(config)
        self.config = config

        colorama_init()

        self.current = None
        if create:
            self.terminal = self.create()

    def validate_lines(self, lines):
        """ validate lines
        """
        if lines < 0 or lines > MAX_LINES:
            raise ValueError(f'lines must be between 1-{MAX_LINES}')

    def validate_config(self, config):
        """ validate progress bar
        """
        if 'progress_bar' in config:
            keys = config['progress_bar'].keys()
            required_keys = ['total', 'count_regex']

            if not all(item in keys for item in required_keys):
                raise ValueError(f'progress_bar config does not contain all the required parameters: {required_keys}')

            if not isinstance(config['progress_bar']['total'], (str, int)):
                raise ValueError('progress_bar.total must be an integer or string')

    def create(self):
        """ return list of dictionaries representing terminal for config
        """
        logger.debug('creating terminal')
        terminal = []
        for index in range(self.lines):
            item = {
                'id': str(index).zfill(self.zfill),
                'text': ''
            }
            if self.config.get('progress_bar'):
                item['count'] = 0
                item['modulus_count'] = 0
                item['total'] = None
            terminal.append(item)
        return terminal

    def get_id_width(self):
        """ return id width
        """
        id_width = self.config.get('id_width', ID_WIDTH)
        if id_width > ID_WIDTH:
            return ID_WIDTH
        return id_width

    def assign_id(self, offset, text):
        """ assign id for offset using id_regex from config
        """
        id_width = self.get_id_width()
        regex_id = self.config['id_regex']
        match_id = re.match(regex_id, text)
        if match_id:
            value = match_id.group('value')
            if self.config.get('id_justify', False):
                if len(value) > id_width:
                    value = f'...{value[-(id_width - 3):]}'
                else:
                    value = value.rjust(id_width)
            self.terminal[offset]['id'] = value
            self.terminal[offset]['id_matched'] = True

    def assign_total(self, offset, text):
        """ assign total for offset using total from config
        """
        if isinstance(self.config['progress_bar']['total'], str):
            regex_total = self.config['progress_bar']['total']
            match_total = re.match(regex_total, text)
            if match_total:
                self.terminal[offset]['total'] = int(match_total.group('value'))
        else:
            self.terminal[offset]['total'] = self.config['progress_bar']['total']

        if self.terminal[offset]['total']:
            self.terminal[offset]['modulus'] = round(self.terminal[offset]['total'] / PROGRESS_BAR_WIDTH)
            # in case total less than progress bar width then lets set total to 1 to avoid divide by zero
            if self.terminal[offset]['modulus'] == 0:
                self.terminal[offset]['modulus'] = 1

    def get_progress_text(self, offset, text):
        """ process progress bar
        """
        if not self.terminal[offset]['total']:
            self.assign_total(offset, text)

        if self.terminal[offset]['count'] == self.terminal[offset]['total']:
            message = self.config.get('progress_bar', {}).get('progress_message', 'Processing complete')
            return message

        regex_count = self.config['progress_bar']['count_regex']
        match_count = re.match(regex_count, text)
        if match_count:
            self.terminal[offset]['count'] += 1
            modulus = self.terminal[offset]['count'] % self.terminal[offset]['modulus']
            if modulus == 0:
                self.terminal[offset]['modulus_count'] = round(round(self.terminal[offset]['count'] / self.terminal[offset]['total'], 2) * PROGRESS_BAR_WIDTH)
                progress = PROGRESS_TICKER * self.terminal[offset]['modulus_count']
                padding = ' ' * (PROGRESS_BAR_WIDTH - self.terminal[offset]['modulus_count'])
                percentage = str(round((self.terminal[offset]['count'] / self.terminal[offset]['total']) * 100)).rjust(3)
                indicator = f"{self.terminal[offset]['count']}/{self.terminal[offset]['total']}"
                return f"Processing |{progress}{padding}| {Style.BRIGHT}{percentage}%{Style.RESET_ALL} {indicator}"
            else:
                if self.terminal[offset]['count'] == self.terminal[offset]['total']:
                    progress = PROGRESS_TICKER * PROGRESS_BAR_WIDTH
                    indicator = f"{self.terminal[offset]['count']}/{self.terminal[offset]['total']}"
                    return f"Processing |{progress}| {Style.BRIGHT}100%{Style.RESET_ALL} {indicator}"

    def write_text(self, offset, text, ignore_progress=False):
        """ write text at offset
        """
        if self.config.get('id_regex') and not self.terminal[offset].get('id_matched', False):
            self.assign_id(offset, text)

        if text == '' or text == self.terminal[offset]['text']:
            # no text or same text to display
            # move to offset but do not print text
            self.write_line(offset, None, None)
            return

        if self.config.get('progress_bar') and not ignore_progress:
            text_to_print = self.get_progress_text(offset, text)
            if not text_to_print:
                return
        else:
            text_to_print = self.sanitize(text)

        id_ = self.terminal[offset]['id']
        id_to_print = f"{Style.BRIGHT + Fore.YELLOW + Back.BLACK}{id_}{Style.RESET_ALL}"
        self.write_line(offset, id_to_print, text_to_print)

    def write_line(self, offset, id_, text):
        """ move to offset and write id_ and text
        """
        move_char = self.get_move_char(offset)
        if text:
            self.terminal[offset]['text'] = text
            print(f'{move_char}{CLEAR_EOL}', end='')
            print(f"{id_}: {text}")
        else:
            print(move_char)
        self.current += 1

    def get_move_char(self, offset):
        """ return move char to move to offset
        """
        move_char = ''
        if offset < self.current:
            move_char = self.move_up(offset)
        elif offset > self.current:
            move_char = self.move_down(offset)
        return move_char

    def move_down(self, offset):
        """ return down to offset
        """
        diff = offset - self.current
        self.current += diff
        return Cursor.DOWN(diff)

    def move_up(self, offset):
        """ return up to offset
        """
        diff = self.current - offset
        self.current -= diff
        return Cursor.UP(diff)

    def write(self, ignore_progress=False):
        """ write lines to screen
        """
        logger.debug('writing terminal')
        if self.current is None:
            self.current = 0
        for offset, item in enumerate(self.terminal):
            self.write_text(offset, item['text'], ignore_progress=ignore_progress)

    def sanitize(self, text):
        """ sanitize text
        """
        if text:
            text = text.splitlines()[0]
            if len(text) > MAX_CHARS:
                return f'{text[0:MAX_CHARS - 3]}...'
        return text

    def cursor(self, hide=True):
        """ show or hide cursor
        """
        if hide:
            print(HIDE_CURSOR, end='')
        else:
            print(SHOW_CURSOR, end='')



#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'mp4ansi',
        version = '0.1.6',
        description = 'A simple ANSI-based terminal emulator that provides multi-processing capabilities.',
        long_description = '[![GitHub Workflow Status](https://github.com/soda480/mp4ansi/workflows/build/badge.svg)](https://github.com/soda480/mp4ansi/actions)\n[![Code Coverage](https://codecov.io/gh/soda480/mp4ansi/branch/main/graph/badge.svg?token=6NTX6LSP7Q)](https://codecov.io/gh/soda480/mp4ansi)\n[![Code Grade](https://www.code-inspector.com/project/20694/status/svg)](https://frontend.code-inspector.com/project/20694/dashboard)\n[![PyPI version](https://badge.fury.io/py/mp4ansi.svg)](https://badge.fury.io/py/mp4ansi)\n\n# mp4ansi #\n\nA simple ANSI-based terminal emulator that provides multi-processing capabilities. MP4ansi will scale execution of a specified function across multiple background processes, where each process is mapped to specific line on the terminal. As the function executes its log messages will automatically be written to the respective line on the terminal. The number of processes along with the arguments to provide each process is specified as a list of dictionaries. The number of elements in the list will dictate the total number of processes to execute (as well as the number of lines in the terminal). The result of each function is written to the respective dictionary element and can be interogated upon completion.\n\nMPansi also supports representing the function execution as a progress bar, you will need to provide an optional config argument containing a dictionary for how to query for the total and count (via regular expressions), see the [examples](https://github.com/soda480/mp4ansi/tree/master/examples) for more detail.\n\nMP4ansi is a subclass of `mpmq`, see [the mpmq PyPi page](https://pypi.org/project/mpmq/) for more information.\n\n### Installation ###\n```bash\npip install mp4ansi\n```\n\n### Examples ###\n\nTo run the samples below you need to install the namegenerator module `pip install namegenerator`.\n\nA simple mp4ansi example:\n```python\nfrom mp4ansi import MP4ansi\nimport uuid, random, namegenerator, time, logging\nlogger = logging.getLogger(__name__)\n\ndef do_work(*args):\n    total = random.randint(200, 600)\n    logger.debug(f\'processing total of {total}\')\n    for _ in range(total):\n        logger.debug(f\'processed {namegenerator.gen()}\')\n        time.sleep(.01)\n    return total\n\nprocess_data = [{} for item in range(8)]\nprint(\'Procesing items...\')\nMP4ansi(function=do_work, process_data=process_data).execute()\nprint(f"Total items processed {sum([item[\'result\'] for item in process_data])}")\n```\n\nExecuting the code above [example1](https://github.com/soda480/mp4ansi/tree/master/examples/example1.py) results in the following:\n![example](https://raw.githubusercontent.com/soda480/mp4ansi/master/docs/images/example1.gif)\n\n**Note** the function being executed `do_work` has no context about multiprocessing or the terminal; it simply perform a function on a given dataset. MP4ansi takes care of setting up the multiprocessing, setting up the terminal, and maintaining the thread-safe queues that are required for inter-process communication.\n\nLet\'s update the example to add an identifer for each process and to show execution as a progress bar. To do this we need to provide additonal configuration via the optional `config` parameter. Configuration is supplied as a dictionary; `id_regex` instructs how to query the identifer from the log messages, `id_justify` will right justify the identifer to make things look nice. For the progress bar, we need to specify `total` and `count_regex` to instruct how to query the total and when to count that an item has been processed respectively. The value for these settings are specified as regular expressions and will match the function log messages, thus we need to ensure our function has log statements for these. If each instance of your function executes on a static data range then you can specify total as an `int`, but in this example the data range is dynamic, i.e. each process will execute on varying data ranges.\n\n```python\nfrom mp4ansi import MP4ansi\nimport uuid, random, namegenerator, time, logging\nlogger = logging.getLogger(__name__)\n\ndef do_work(*args):\n    pid = str(uuid.uuid4())\n    logger.debug(f\'processor id {pid[0:random.randint(8, 30)]}\')\n    total = random.randint(200, 600)\n    logger.debug(f\'processing total of {total}\')\n    for _ in range(total):\n        logger.debug(f\'processed {namegenerator.gen()}\')\n        time.sleep(.01)\n    return total\n\nprocess_data = [{} for item in range(8)]\nconfig = {\n    \'id_regex\': r\'^processor id (?P<value>.*)$\',\n    \'id_justify\': True,\n    \'progress_bar\': {\n        \'total\': r\'^processing total of (?P<value>\\d+)$\',\n        \'count_regex\': r\'^processed (?P<value>.*)$\'}}\nprint(\'Procesing items...\')\nMP4ansi(function=do_work, process_data=process_data, config=config).execute()\nprint(f"Total items processed {sum([item[\'result\'] for item in process_data])}")\n```\n\nExecuting the code above [example2](https://github.com/soda480/mp4ansi/tree/master/examples/example2.py) results in the following:\n![example](https://raw.githubusercontent.com/soda480/mp4ansi/master/docs/images/example2.gif)\n\nMore [examples](https://github.com/soda480/mp4ansi/tree/master/examples) are included to demonstrate the mp4ansi package. To run the examples, build the Docker image and run the Docker container using the instructions described in the [Development](#development) section.\n\nTo run the example scripts within the container:\n\n```bash\npython examples/example#.py\n```\n\n### Development ###\n\nClone the repository and ensure the latest version of Docker is installed on your development server.\n\n\nBuild the Docker image:\n```sh\ndocker image build \\\n-t \\\nmp4ansi:latest .\n```\n\nRun the Docker container:\n```sh\ndocker container run \\\n--rm \\\n-it \\\n-v $PWD:/mp4ansi \\\nmp4ansi:latest \\\n/bin/sh\n```\n\nExecute the build:\n```sh\npyb -X\n```\n',
        long_description_content_type = 'text/markdown',
        classifiers = [
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Environment :: Other Environment',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: POSIX :: Linux',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3.6',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: System :: Networking',
            'Topic :: System :: Systems Administration'
        ],
        keywords = '',

        author = 'Emilio Reyes',
        author_email = 'soda480@gmail.com',
        maintainer = '',
        maintainer_email = '',

        license = 'Apache License, Version 2.0',

        url = 'https://github.com/soda480/mp4ansi',
        project_urls = {},

        scripts = [],
        packages = ['mp4ansi'],
        namespace_packages = [],
        py_modules = [],
        entry_points = {},
        data_files = [],
        package_data = {},
        install_requires = [
            'mpmq',
            'colorama'
        ],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '',
        obsoletes = [],
    )
